#ifndef SPLINESSPLINE_H
#define SPLINESSPLINE_H

// ----------------------------------------------------------------------
//
// Spline.h
//
// Classes embodying the Spline concept 
//
// The generic d-dimensional Spline<d> class template
// Specialization to Spline<1> in one dimension 
// Specialization to Spline<2> 
// ... up to ...
// Specialization to Spline<6> in six dimensions
//
// ----------------------------------------------------------------------

#include "splines/Grid.h"
#include "splines/DeltaFunction.h"

#define EMBEDDING

#ifdef EMBEDDING
#include "cpp0x/functional"
#endif
#include <vector>
#include <iostream>

namespace splines {

// The following template will be defined in this header:

template <unsigned int d> class Spline;

// The following specializations will be defined in this header:

template <> class Spline<1>;
template <> class Spline<2>;
template <> class Spline<3>;
template <> class Spline<4>;
template <> class Spline<5>;
template <> class Spline<6>;

// The following allows use of whatever syntax defines a function of
// a function -- either boost:function (in the original) or std::function 
// (in c++0x).  

#ifdef EMBEDDING
typedef std::function<double (double)> Function_1;
typedef std::function<double (double,double)> Function_2;
#else
typedef double  Function_1(double);
typedef double  Function_2(double,double);
#endif

// ==========
// Spline <1> 
// ==========

template <>
class Spline<1> {

public:
 
  // Constructors and assignment

  // From an ordinary function
  Spline ( double f(double x), const Grid<1> & g );

  // From a function wrapping an indexed table
  Spline ( double f(int x_index), const Grid<1> & g );

  // From an indexed table
  Spline ( std::vector<double> const & suppliedValues, const Grid<1> & g );

  // From a functor
  template <class F>		// F must have the following property:
  Spline ( const F & f, 	// f(double x) must return a double.
	   const Grid<1> & g )	: grid(g)
	   			, specifiedByTable(false)
#ifdef EMBEDDING
	   			, embedding()
#endif
                                , isDeltaFunction(false)
				{  captureValues ( f );
				   computeDerivatives  ( ); }
  // TODO - Spline<1> s ( std::sin, grid ) does not compile due to 
  // overloading ambiguities.
  
  // From a DeltaFunction
  explicit Spline ( DeltaFunction<1> const & d );

  // From a stream containing the saved state of a Spline<1> 
  explicit Spline (std::istream & is) 
                                : isDeltaFunction(false)
                                { setTables(is); }

  // From pre-computed values and derivatives
  Spline ( const Grid<1> & g, std::vector<double> const &, 
                              std::vector<double> const & derivs);

  // Use as generic spline
  Spline ( double f(std::vector<double> const & v), const Grid<1> & g );

  // Use as generic spline defined by table
  Spline (  double f(std::vector<int> const & indicies), const Grid<1> & g );

  // Copy, swap, and def ctors
  // copy and operator= are fine with compiler-generated versions
  void swap (Spline<1> & s);
  Spline() { }

#ifdef EMBEDDING 
  // Embedded splines:  In each of these cases, if a functor is used as 
  // the Outer type, the supplied outerF must remain valid for the 
  // lifetime of the spline.
				   
  // Embedded spline from an ordinary function or a functor
  template <class F, class Outer> // Both F and Outer must have the following 
  Spline ( const F & f, 	  // property: f(double x) must return a double.
	   const Grid<1> & g,	  
	   const Outer & outerF ) 
	   : grid(g)
	   , specifiedByTable(false)
#ifdef EMBEDDING
	   , embedding( outerF )
#endif
           , isDeltaFunction(false)
				{  captureValues ( f );
				   computeDerivatives  ( ); }
	   			  

  // Embedded spline from a function wrapping an indexed table
  template <class Outer> 	  
  Spline ( double f(int x_index), 
           const Grid<1> & g,     
	   const Outer & outerF ) 
	   : grid(g)
	   , specifiedByTable(true)
#ifdef EMBEDDING
	   , embedding( outerF )
#endif
           , isDeltaFunction(false)
				{  captureValues ( f );
				   computeDerivatives  ( ); }

  // Embedded spline from a stream containing the saved state of a Spline<1> 
  template <class Outer>          
  explicit 			  
  Spline (std::istream & is, const Outer & outerF ) 
                                : embedding (outerF)
                                , isDeltaFunction(false)
  				{  setTables(is);  }

  // Embedded spline from pre-computed values and derivatives
  template <class Outer> 
  Spline ( const Grid<1> & g, 
           std::vector<double> const & vals, 
           std::vector<double> const & derivs,
           const Outer & outerF )
	   : grid(g)
           , values(vals)
           , derivatives(derivs)
#ifdef EMBEDDING
	   , embedding( outerF )
#endif
           , isDeltaFunction(false)
{
}

  // Use as generic embedded spline 
  Spline ( double f(std::vector<double> const & v), const Grid<1> & g,
  	   double outerF(std::vector<double> const & v ) );
				   
  // Use as generic embedded spline defined by table of values
  Spline ( double f(std::vector<int> const & indicies), const Grid<1> & g, 
  	   double outerF(std::vector<double> const & v ) ); 

  // Use as generic embedded spline from pre-computed values and derivatives
  Spline ( const Grid<1> & g, std::vector<double> const & vals, 
                              std::vector<double> const & derivs,
  	   double outerF(std::vector<double> const & v ) );
#endif

  // Invoking the Spline

  double operator() (double x) const;				
  double operator() (const std::vector<double> & v) const;	

  // Equality comparison
  
  bool operator== ( Spline<1> const & rhs ) const;
  bool operator!= ( Spline<1> const & rhs ) const {return !(operator==(rhs));}
  
// #error Correct Splines header used
  
  // Capturing the spline data for use in other jobs
  
  void 
  save (std::vector<double> & values, std::vector<double> & derivatives) const;

  std::ostream & put (std::ostream & os) const; 

  // Specialized operations:  Convolution, integration, optimization
  
  // "Smear" in the range of this spline, with a unit-normalized g(p) 
  // by doing s(q) = integral [f(q-p)g(p)] dp
  Spline<1> convolution(Spline<1> const & g) const;
  Spline<1> convolution(Function_1 const & g, double g_min, double g_max) const;
  // TODO - convolution involving embedded splines is not implemented.
  
  // Definite and indefinite integrals
  double integrate(double low, double high) const;
  Spline<1> indefiniteIntegral(double lowerLimit) const ;
        // if I = s.indefiniteIntegral(x), then I(x) = 0 and
        // I(b) - I(a) = s.integrate(a,b) -- at least to third order 
  // TODO - integration involving embedded splines is not implemented.

  // Maximum and minimum (restricted to within the grid)
  // note that maximum() and minimum() return a value of x along the grid,
  // not values of the Spline at that point x.
  double maximum() const;  
  double minimum() const;
  // TODO - extremum involving embedded splines is not implemented.
    
private:

  Grid<1> grid;
  std::vector<double> values;
  std::vector<double> derivatives;

  bool specifiedByTable;
  Function_1 embedding;
  double slope0;
  double slopeN;
  bool isDeltaFunction;
  double deltaFunctionLocation;
  double deltaFunctionStrength;

  void captureValues ( double f(double) );		
  void captureValues ( double f(int x_index) );	
  void captureValues ( std::vector<double> const & suppliedValues );	
  void captureValues ( double f(std::vector<double> const & v) );  
  void captureValues ( double f(std::vector<int> const & indicies) );  
  void setTables (std::istream & is);
   
  template <class F> void captureValues (const F & f) {
    unsigned int n = grid.nPoints();
    values.clear(); values.reserve( n );
    double x;
    for (unsigned int index = 0; index < n; ++index) {
      x = grid[index];
      values.push_back ( f(x) );
    }
    specifiedByTable = false;
    const double smallFraction = .01;
    double xs = grid[0] * (1-smallFraction) + grid[1] * smallFraction;
    slope0 = (f(xs) - values[0])/(xs - grid[0]); 
    xs = grid[n-1] * (1-smallFraction) + grid[n-2] * smallFraction; 
    slopeN = (values[n-1] - f(xs))/(grid[n-1] - xs);  
  }

  void computeDerivatives();					
  static
  std::vector<double> secondDerivatives (const std::vector<double> & x,
                   			 std::vector<double> const & vals,
                   			 double a0, double s0,
                   			 double bN, double sN);

  inline double f(double x) const {return operator()(x);}
  
  Spline<1> deltaFunctionConvolution (Spline<1> const & g) const;

}; // Spline<1>


// ==========
// Spline <2> 
// ==========

template <>
class Spline<2> {

public:
 
  // Constructors and assignment

  // From an ordinary function
  Spline ( double f(double x, double y), const Grid<2> & g );

  // From a function wrapping an indexed table
  Spline ( double f(int x_index, int y_index), const Grid<2> & g );

  // From a functor
  template <class F>		// F must have the following property:
  Spline ( const F & f, 	// f(double x, double y) must return a double.
	   const Grid<2> & g )	: grid(g)
	   			, specifiedByTable(false)
#ifdef EMBEDDING
	   			, embedding()
#endif
				{  captureValues ( f );
				   computeDerivatives  ( ); }

  // From a stream containing the saved state of a Spline<1> 
  explicit Spline (std::istream & is) { setTables(is); }

  // From pre-computed values and derivatives
  Spline ( const Grid<2> & g, std::vector<double> const & vals, 
                              std::vector<double> const & derivs);

  // Use as generic spline
  Spline ( double f(std::vector<double> const & v), const Grid<2> & g );

  // Use as generic spline defined by table
  Spline (  double f(std::vector<int> const & indicies), const Grid<2> & g );

  // Copy, Swap, and def ctors
  Spline (const Spline<2> & s);
  Spline<2> & operator= (const Spline<2> & s);
  void swap (Spline<2> & s);
  Spline() { }
 
#ifdef EMBEDDING
  // Embedded splines:  In each of these cases, if a functor is used as 
  // the Outer type, the supplied outerF must remain valid for the 
  // lifetime of the spline.
				   
  // Embedded spline from an ordinary function or a functor
  template <class F, class Outer> // Both F and Outer must have the following 
  Spline ( const F & f, 	  // property: f(double x, double y) 
	   const Grid<2> & g,	  // must return a double. 
	   const Outer & outerF ) 
	   : grid(g)
	   , specifiedByTable(false)
	   , embedding( outerF )
				{  captureValues ( f );
				   computeDerivatives  ( ); }
	   			  
  // Embedded spline from a function wrapping an indexed table
  template <class Outer> 	  
  Spline ( double f(int x_index, int y_index), 
           const Grid<2> & g,     
	   const Outer & outerF ) 
	   : grid(g)
	   , specifiedByTable(true)
	   , embedding( outerF )
				{  captureValues ( f );
				   computeDerivatives  ( ); }

  // Embedded spline from a stream containing the saved state of a Spline<1> 
  template <class Outer>          
  explicit 			  
  Spline (std::istream & is, const Outer & outerF ) : embedding (outerF)
  				{  setTables(is);  }

  // Embedded spline from pre-computed values and derivatives
  template <class Outer> 
  Spline ( const Grid<2> & g, 
           double *vals, 
           double *derivs,
           const Outer & outerF )
	   : grid(g)
	   , embedding( outerF )
	   			{ values = vals;
				  derivatives = derivs; }

  // Use as generic embedded spline 
  Spline ( double f(std::vector<double> const & v), const Grid<2> & g,
  	   double outerF(std::vector<double> const & v ) );
				   
  // Use as generic embedded spline defined by table of values
  Spline ( double f(std::vector<int> const & indicies), const Grid<2> & g, 
  	   double outerF(std::vector<double> const & v ) ); 

  // Use as generic embedded spline from pre-computed values and derivatives
  Spline ( const Grid<2> & g, double *vals, double *derivs,
  	   double outerF(std::vector<double> const & v ) );
#endif

  // Invoking the Spline

  double operator() (double x, double y) const;				
  double operator() (const std::vector<double> & v) const;	

  // Equality comparison
  
  bool operator== ( Spline<2> const & rhs ) const;
  bool operator!= ( Spline<2> const & rhs ) const {return !(operator==(rhs));}
  
  
  // Capturing the spline data for use in other jobs
  
  void 
  save (std::vector<double> & values, std::vector<double> & derivatives) const;

  std::ostream & put (std::ostream & os) const; 

  
private:

  Grid<2> grid;
  int nx;
  int ny;
  std::vector<double> values_vector;
  std::vector<double> derivatives_vector;
  double * values;
  double * derivatives;
  double * dx;
  double * dy;
  double * dxy;
  bool specifiedByTable;
  Function_2 embedding;
  std::vector<double> x_slope0;
  std::vector<double> x_slopeN;
  std::vector<double> y_slope0;
  std::vector<double> y_slopeN;
  double dxy00;
  double dxy0N;
  double dxyN0;
  double dxyNN;

  void captureValues ( double f(double, double) );		
  void captureValues ( double f(int x_index, int y_index) );	
  void captureValues ( double f(std::vector<double> const & v) );  
  void captureValues ( double f(std::vector<int> const & indicies) );  
  void setTables (std::istream & is);
   
  template <class F> void boundaryDerivatives(F f) {
    const double smallFraction = .01;
    x_slope0.reserve(ny);
    y_slope0.reserve(nx);
    x_slopeN.reserve(ny);
    y_slopeN.reserve(nx);
    std::vector<double> gx = grid.points(0);
    std::vector<double> gy = grid.points(1);
    // Slopes in the x direction for all values of y
    double xs0 = gx[0]    * (1-smallFraction) + gx[1]    * smallFraction;
    double xsN = gx[nx-1] * (1-smallFraction) + gx[nx-2] * smallFraction; 
    double kx0 = 1.0 / (xs0 - gx[0]);    // denominator in derivative at left
    double kxN = 1.0 / (gx[nx-1] - xsN); // denominator in derivative at right
    const int vx0 = 0;
    const int vxN = (nx-1)*ny;
    for ( unsigned int iy = 0; iy < ny; ++iy ) {
      double temp = (f(xs0,gy[iy]) - values[vx0+iy]);
      x_slope0.push_back ( (f(xs0,gy[iy]) - values[vx0+iy])*kx0 ); 
      x_slopeN.push_back ( (values[vxN+iy] - f(xsN,gy[iy]))*kxN );  
    } 
    // Slopes in the y direction for all values of x
    double ys0 = gy[0]    * (1-smallFraction) + gy[1]    * smallFraction;
    double ysN = gy[ny-1] * (1-smallFraction) + gy[ny-2] * smallFraction; 
    double ky0 = 1.0 / (ys0 - gy[0]);    // denominator in derivative at bottom
    double kyN = 1.0 / (gy[ny-1] - ysN); // denominator in derivative at top
    for ( int ix = 0, ixny=0; ix < nx; ++ix, ixny += ny ) {
      y_slope0.push_back ( (f(gx[ix],ys0) - values[ixny])*ky0 ); 
      y_slopeN.push_back ( (values[ixny+ny-1] - f(ysN,gy[ny-1]))*kxN );  
    } 
    // Mixed second derivatives at the corners
    dxy00 =(f(xs0,ys0) - f(gx[0],ys0) - f(xs0,gy[0]) - f(gx[0],gy[0]))*kx0*ky0; 
    dxy0N =(f(xs0,ysN)-f(gx[0],ysN)-f(xs0,gy[ny-1])-f(gx[0],gy[ny-1]))*kx0*kyN; 
    dxyN0 =(f(xsN,ys0)-f(gx[nx-1],ys0)-f(xsN,gy[0])-f(gx[nx-1],gy[0]))*kxN*ky0; 
    dxyNN =(f(xsN,ysN)-f(gx[nx-1],ysN)-f(xsN,gy[ny-1])-f(gx[nx-1],gy[ny-1]))
    								      *kxN*kyN; 
  }
  
  template <class F> void captureValues (const F & f) {
    std::vector<int> shape = grid.shape();
    nx = shape[0];
    ny = shape[1];
    values_vector.clear(); values_vector.reserve( nx*ny );
    std::vector<double> gx = grid.points(0);
    std::vector<double> gy = grid.points(1);
    double x;
    double y;
    for (unsigned int ix = 0; ix < nx; ++ix) {
     x = gx[ix];
     for (unsigned int iy = 0; iy < ny; ++iy) {
       y = gy[iy];
       values_vector.push_back ( f(x,y) );
     }
    }
    values = &values_vector[0];
    specifiedByTable = false;
    boundaryDerivatives(f);
  }

  void computeDerivatives();
  void computeLineOfDerivatives( std::vector<double> const & x,
  				 std::vector<double> const & values,
  				 double * dest,
				 int stride );					
  void computeLineOfDerivatives( std::vector<double> const & x,
  				 std::vector<double> const & values,
  				 double * dest,
				 int stride,
				 double slope0,
				 double slopeN );					
  static
  std::vector<double> secondDerivatives (const std::vector<double> & x,
                   			 const double * vals,
                   			 double a0, double s0,
                   			 double bN, double sN);

}; // Spline<2>








#ifdef NOTYET
#error this should not be here

// ==========
// Spline <3> 
// ==========

template <>
class Spline<3> {

public:

  static const unsigned int d = 3;

  // Constructors and assignment

  template <class F>		// F must have the following property:
  Spline ( const F & f, 	// double f(double x, double y, double z) 
	   Grid<d> g,		// Either an ordinary function or a 
				// functor object is acceptable.
	   ApproximationLevel approx = CUBIC_SPLINE );

  Spline ( double f(int x_index, int y_index, int z_index),
	   Grid<d> g, 
	   ApproximationLevel approx = CUBIC_SPLINE );

  template <class F>		
  Spline ( const F & f, Grid<d> g, SplineControl<d> c);	

  Spline ( double f(double, double, double), Grid<d> g, 
					const SplineControl<d> c);	
  Spline ( double f(int x_index, int y_index, int z_index), Grid<d> g,   
					const SplineControl<d> c);	

  Spline (const Spline<d> & s);
  Spline<d> & operator= (const Spline<d> & s);
  void swap (Spline<d> & s);

  Spline()  { }
  ~Spline() { }

  // Invoking the Spline

  double operator() (double x, double y, double z); 

private:

  template <class F> void captureValues         (const F & f);
  template <class F> void captureValuesByIndices(const F & f);

}; // Spline<3>



// ==========
// Spline <4> 
// ==========

template <>
class Spline<4> {

public:

  static const unsigned int d = 4;

  // Constructors and assignment

  template <class F>		// F must have the following property:
  Spline ( const F & f, 	// double f(w, x, y, z) // w,x,y,z are doubles
	   Grid<d> g,		// Either an ordinary function or a 
				// functor object is acceptable.
	   ApproximationLevel approx = CUBIC_SPLINE );

  Spline ( double f(int w_index, int x_index, int y_index, int z_index),
	   Grid<d> g, 
	   ApproximationLevel approx = CUBIC_SPLINE );

  template <class F>		
  Spline ( const F & f, Grid<d> g, SplineControl<d> c);	

  Spline ( double f(double, double, double, double), Grid<d> g, 
					const SplineControl<d> c);	
  Spline ( double f(int w_index, int x_index, int y_index, int z_index), 
			   Grid<d> g,   const SplineControl<d> c);	


  Spline<d> & operator= (const Spline<d> & s);
  void swap (Spline<d> & s);

  Spline()  { }
  ~Spline() { }

  // Invoking the Spline

  double operator() (double w, double x, double y, double z); 

private:

  template <class F> void captureValues         (const F & f);
  template <class F> void captureValuesByIndices(const F & f);

}; // Spline<4>


// ==========
// Spline <5> 
// ==========

template <>
class Spline<5> {

public:

  static const unsigned int d = 5;

  // Constructors and assignment

  template <class F>		// F must have the following property:
  Spline ( const F & f, 	// double f(v, w, x, y, z) // args are doubles
	   Grid<d> g,		// Either an ordinary function or a 
				// functor object is acceptable.
	   ApproximationLevel approx = CUBIC_SPLINE );

  Spline ( double f(int v_index, int w_index, 
		    int x_index, int y_index, int z_index),
	   Grid<d> g, 
	   ApproximationLevel approx = CUBIC_SPLINE );

  template <class F>		
  Spline ( const F & f, Grid<d> g, SplineControl<d> c);	

  Spline ( double f(double, double, double, double, double), Grid<d> g, 
					const SplineControl<d> c);	
  Spline ( double f(int v_index, int w_index, int x_index, 
			int y_index, int z_index), 
			   Grid<d> g,   const SplineControl<d> c);	


  Spline (const Spline<d> & s);
  Spline<d> & operator= (const Spline<d> & s);
  void swap (Spline<d> & s);

  Spline()  { }
  ~Spline() { }

  // Invoking the Spline

  double operator() (double v, double w, double x, double y, double z); 

private:

  template <class F> void captureValues         (const F & f);
  template <class F> void captureValuesByIndices(const F & f);

}; // Spline<5>


// ==========
// Spline <6> 
// ==========

template <>
class Spline<6> {

public:

  static const unsigned int d = 6;

  // Constructors and assignment

  template <class F>		// F must have the following property:
  Spline ( const F & f, 	// double f(u, v, w, x, y, z) // all doubles
	   Grid<d> g,		// Either an ordinary function or a 
				// functor object is acceptable.
	   ApproximationLevel approx = CUBIC_SPLINE );

  Spline ( double f (int u_index, int v_index, int w_index, 
	             int x_index, int y_index, int z_index),
	   Grid<d> g, 
	   ApproximationLevel approx = CUBIC_SPLINE );

  template <class F>		
  Spline ( const F & f, Grid<d> g, SplineControl<d> c);	

  Spline ( double f(double, double, double, double, double, double), Grid<d> g, 
					const SplineControl<d> c);	
  Spline ( double f(int u_index, int v_index, int w_index, 
		    int x_index, int y_index, int z_index), 
			   Grid<d> g,   const SplineControl<d> c);	

  Spline (const Spline<d> & s);
  Spline<d> & operator= (const Spline<d> & s);
  void swap (Spline<d> & s);

  Spline()  { }
  ~Spline() { }

  // Invoking the Spline

  double operator()(double u, double v, double w, double x, double y, double z);

private:

  template <class F> void captureValues         (const F & f);
  template <class F> void captureValuesByIndices(const F & f);

}; // Spline<6>


#endif // NOTYET

// ==========
// Spline <d> 
// ==========

template <unsigned int d>
class Spline {

public:

  // Constructors and assignment

  // From an ordinary function
  Spline ( double f(std::vector<double> v), const Grid<d> & g );

  // From a function wrapping an indexed table
  Spline ( double f(std::vector<int> indices), const Grid<d> & g );
  Spline ( double f(std::vector<unsigned int> indices), const Grid<d> & g );

  // From a functor
  template <class F>		// F must have the following property:
  Spline ( const F & f, 	// f(std::vector<double> v) must return a double
	   const Grid<d> & g );	

  // From a stream containing the saved state of a Spline<d> 
  explicit Spline (std::istream & is);

  // Embedded splines:  In each of these cases, if a functor is used as 
  // the Outer type, the supplied outerF must remain valid for the 
  // lifetime of the spline.
				   
  // Embedded spline from an ordinary function or a functor
  template <class F, class Outer> // Both F and Outer must have the following 
  Spline ( const F & f, 	 // property: 
	   const Grid<d> & g,	// f(std::vector<double> v) must return a double
	   const Outer & outerF );
	   			  
  // Embedded spline from a function wrapping an indexed table
  template <class Outer> 	  
  Spline ( double f(std::vector<int> indices), 
           const Grid<d> & g,     
	   const Outer & outerF );
  template <class Outer> 	  
  Spline ( double f(std::vector<unsigned int> indices), 
           const Grid<d> & g,     
	   const Outer & outerF );

  // Embedded spline from a stream containing the saved state of a Spline<d> 
  template <class Outer>          
  explicit 			  
  Spline (std::istream & is, const Outer & outerF );

  // Copy and assignment
  Spline (const Spline<d> & s);
  Spline<d> & operator= (const Spline<d> & s);
  void swap (Spline<d> & s);

  Spline()  { }
  ~Spline() { }

  // Invoking the Spline

  double operator() (const std::vector<double> & v);

private:

  Grid<d> grid;
  std::vector<double> values;  // TODO - the nature of this MAY change

  

  template <class F> void captureValues         (const F & f);

}; // Spline<d>

// General helper functions:

// ==========================================================================
//
//	Inline Implementations
//
// ==========================================================================

// ----------------------------------------------------------------------
//
// Common template code for the specializations Spline<1> - Spline<6>
// and template code implementing Spline<d>.
//
// We attempt to minimize the bulk of templated code by factoring out
// as much as possible into code suitable for a .cc file.
//
// ----------------------------------------------------------------------

// ---------------
// Inlined helpers
// ---------------

inline static bool
advanceVector ( std::vector<int> & c, const std::vector<int> & gshape ) {
  for ( int n = c.size()-1; n>=0; --n ) {
    ++c[n];
    if ( c[n] >= gshape[n] ) { c[n] = 0; continue; }
    return true;
  }
  return false;
}

inline static int gshapeProduct ( const std::vector<int> & gshape ) {
  int product = 1;
  for ( std::vector<int>::const_iterator it = gshape.begin(); 
	it != gshape.end(); ++it ) product *= *it;
  return product;
}


// ==========
// Spline <1> 
// ==========


#ifdef MOVED_TO_CC
template <class F>
void
Spline<1>::captureValues (const F & f) {
  
  int n = grid.nPoints(); 
  values.clear(); values.reserve( n );
  double x;
  for (unsigned int index = 0; index < n; ++index) {
    x = grid[index];
    values.push_back ( f(x) );
  }
}

template <class F>
Spline<1>::Spline ( const F & f, const Grid<1> & g, int approx ) : 
					grid(g), splineApprox(approx)
{
  captureValues ( f );
  computeDerivatives  ( );
}
#endif 

// ==========
// Spline <d> 
// ==========

template <unsigned int d>
template <class F>
void
Spline<d>::captureValues (const F & f) {
  std::vector<int> indices(d,0);
  std::vector<int> gshape = grid.shape();
  // #TODO Grid<d> needs vector<int> shape()
  // #TODO Grid<d> needs Coordinates<d> operator[]()
  values.clear(); values.reserve( gshapeProduct(gshape) );
  do {
    values.push_back ( f(grid[indices]));
  }
  while ( advanceVector(indices, gshape) );
}

// ===
// I/O
// ===

template<unsigned int d>
std::ostream & operator<< (std::ostream& os, splines::Spline<d> const & s)
{ 
  return s.put(os);
}

} // end of namespace splines


#endif  /* SPLINESSPLINE_H */
